/*******************************************************************************
 *
 * FILENAME:     main.cpp
 * DESCRIPTION:  main test file
 *
 * AUTHOR:       Jun Liu (jun@cs.man.ac.uk)
 * INSTITUTION:  Advanced Interfaces Group,
 *               Department of Computer Science,
 *               the University of Manchester, United Kingdom
 * DATE:         August 2006
 *
 *******************************************************************************/

#include "GaussianNoise.h"
#include "RotationMatrix.h"
#include "ProjectionMatrix.h"
#include "CalibrationMatrix.h"
#include "PlyReader.h"

#include <mvl/HomgPoint3D.h>
#include <mvl/HomgPoint2D.h>
#include <iostream>
#include <cstdlib>
#include <fstream>
#include <string>
#include <algorithm>


using namespace std;

const double POINTS_SPACE                = 10.0;
const double CAMERA_RADIUS               = 100.0;
const double FOCAL                       = 1400;
const double PRINCIPAL_X                 = 400;
const double PRINCIPAL_Y                 = 300;
const double VARIANCE                    = 1.0;

const int NUM_POINTS                     = 100;
const int NUM_CAMERAS                    = 36;


void generate3DPoints(vector<HomgPoint3D> & points,
                      int    NumOfPoints,
                      double MaxX =  POINTS_SPACE,
                      double MinX = -POINTS_SPACE,
                      double MaxY =  POINTS_SPACE,
                      double MinY = -POINTS_SPACE,
                      double MaxZ =  POINTS_SPACE,
                      double MinZ = -POINTS_SPACE)
    
{
    assert (NumOfPoints > 0);
    points.resize(NumOfPoints);
    
    double rangeX = MaxX - MinX;
    double rangeY = MaxY - MinY;
    double rangeZ = MaxZ - MinZ;
    
    for (int i=0; i<NumOfPoints; ++i)
    {
        double x = (rand()/(RAND_MAX + 1.0)) * rangeX + MinX;
        double y = (rand()/(RAND_MAX + 1.0)) * rangeY + MinY;
        double z = (rand()/(RAND_MAX + 1.0)) * rangeZ + MinZ;
        
        points[i] = HomgPoint3D(x,y,z);
            // points[i] = HomgPoint3D(0,0,0);
        
    }
}



void scalePoints(vector<HomgPoint3D> & points)
{
    vector<double> Xs, Ys, Zs;

    int size = points.size();

    for (int i=0; i<size; ++i)
    {
        Xs.push_back(points[i].x());
        Ys.push_back(points[i].y());
        Zs.push_back(points[i].z());
    }

    sort(Xs.begin(), Xs.end());
    sort(Ys.begin(), Ys.end());
    sort(Zs.begin(), Zs.end());

    HomgPoint3D median(Xs[size/2], Ys[size/2],Zs[size/2]);

    vector <double> distance;

    for (int i=0; i<size; ++i)
    {
        double dx = median.x() - Xs[i];
        double dy = median.y() - Ys[i];
        double dz = median.z() - Zs[i];
        double dd = sqrt(dx*dx + dy*dy + dz*dz);
        distance.push_back(dd);
    }
    
    sort (distance.begin(),distance.end());

    double median_distance = distance[size/2];

    for (int i=0; i<size; ++i)
    {
        points[i] = HomgPoint3D((points[i].x() - median.x())*  POINTS_SPACE / median_distance,
                                (points[i].y() - median.y())*  POINTS_SPACE / median_distance,
                                (points[i].z() - median.z())*  POINTS_SPACE / median_distance);
    }
    
    
    
}


// load 3D points from PLY file
void load3DPoints(char * filename,
                  vector<HomgPoint3D> & points)
{
    string file (filename);
    PlyFile ply (file);

    
    for (unsigned int i=0; i<ply.elements.size(); ++i)
        if (ply.elements[i].name=="vertex")
        {
            int numPoints = ply.elements[i].numElements;
            points.resize(numPoints);
            
            
            int PropertySize = ply.elements[i].properties.size();
            
            
            int count = 0;
            for (int j=0; j<PropertySize; ++j)
            {
                    // cout << ply.elements[i].properties[j]->type << " "
                    // << ply.elements[i].properties[j]->name << " "
                    // << ply.elements[i].properties[j]->data.size() << endl;

                if (ply.elements[i].properties[j]->name == "x")
                {
                    for (int k=0; k<numPoints; ++k)
                        points[k].set(atof(ply.elements[i].properties[j]->data[k].c_str()),
                                      points[k].y(),
                                      points[k].z());
                    ++count;
                }

                if (ply.elements[i].properties[j]->name == "y")
                {
                    for (int k=0; k<numPoints; ++k)
                        points[k].set(points[k].x(),
                                      atof(ply.elements[i].properties[j]->data[k].c_str()),
                                      points[k].z());
                    ++count;
                }
                
                if (ply.elements[i].properties[j]->name == "z")
                {
                    for (int k=0; k<numPoints; ++k)
                        points[k].set(points[k].x(),
                                      points[k].y(),
                                      atof(ply.elements[i].properties[j]->data[k].c_str()));
                    ++count;
                }

            }
            
            assert (count == 3);
            
        }
    
    
    scalePoints(points);
}

    


// pure rotation
void generateCamerasOrbital(int                      NumOfCameras,
                            vector<PMatrix>        & cameras,
                            vector<vnl_double_3x3> & R,
                            vector<vnl_double_3>   & t,
                            vector<vnl_double_3x3> & K)
{
    assert(NumOfCameras > 0);
    cameras.resize(NumOfCameras);
    R.resize(NumOfCameras);
    t.resize(NumOfCameras);
    K.resize(NumOfCameras);
    
    
    double step = 2*M_PI / static_cast<double>(NumOfCameras);
    
    
    for (int i=0; i<NumOfCameras; ++i)
    {
        double rotation = step * i;
        
  	R[i] = RotationMatrix::fromEulerAngles(0, rotation, 0);
            //R[i] = RotationMatrix::fromEulerAngles(rotation/50.0, rotation, 0);
            // R[i] = RotationMatrix::fromEulerAngles(0, 0, 0);

        t[i] = vnl_double_3(  CAMERA_RADIUS * sin(rotation),
                              0.0,
                              -CAMERA_RADIUS * cos(rotation));

            // t[i] = vnl_double_3 (0,0,-CAMERA_RADIUS - 5 * i);
        
        
        K[i] = CalibrationMatrix::generate(FOCAL, FOCAL, PRINCIPAL_X, PRINCIPAL_Y);
        cameras[i] = ProjectionMatrix::generate(K[i], R[i], t[i]);
    }
}



int main (int argc, char * argv[])
{
    if (argc != 4)
    {
        cerr << "Usage: " << argv[0] << " in_PLYFile out_GroundTruthFile out_MeasurementFile" << endl;
        exit(1);
    }
    
    vector<HomgPoint3D> points;
    load3DPoints(argv[1], points);
        // generate3DPoints(points,NUM_POINTS);
    
        
    ofstream groundtruthfile(argv[2]);
    if (!groundtruthfile.is_open())
    {
        cerr << "Error opening file " << argv[2] << endl;
        exit(1);
    }
    
    ofstream measurementfile(argv[3]);
    if (!measurementfile.is_open())
    {
        cerr << "Error opening file " << argv[3] << endl;
        exit(1);
    }
    
    
    
    int NumOfCameras  = NUM_CAMERAS;
    int NumOfPoints   = points.size();
    
    measurementfile << static_cast<int>(2 * PRINCIPAL_X) << endl;
    measurementfile << static_cast<int>(2 * PRINCIPAL_Y) << endl;
    measurementfile << NumOfCameras << endl;
    measurementfile << NumOfPoints  << endl;
    
    
    
        // generate3DPoints(points, NumOfPoints);
    
    vector<PMatrix> cameras;
    vector<vnl_double_3x3> R, K;
    vector<vnl_double_3>   t;
    
    generateCamerasOrbital(NumOfCameras, cameras, R, t, K);
    

    groundtruthfile << NumOfCameras << endl ;
    groundtruthfile << NumOfPoints  << endl << endl;
    
    for (int i=0; i<NumOfCameras; ++i)
    {
        groundtruthfile << cameras[i] << endl;
        groundtruthfile << K[i] << endl;
        groundtruthfile << R[i] << endl;
        groundtruthfile << t[i] << endl << endl;
    }
    

    for (int i=0; i<NumOfPoints; ++i)
    {
        groundtruthfile << NumOfCameras << " ";
        measurementfile << NumOfCameras << " ";
        
        for (int j=0; j<NumOfCameras; ++j)
        {
            
            HomgPoint2D p2d = cameras[j].project(points[i]);


                // if (cameras[j].is_behind_camera(points[i]))
                // cout << "behind" << endl;
                // else
                // cout << "front" << endl;
            

            p2d.rescale_w();
            
            double x = p2d.x();
            double y = p2d.y();
            
            groundtruthfile << j << " " << x << " " << y << " " ;
            
            
            HomgPoint2D noisy (GaussianNoise::generateGaussianNoise(x,VARIANCE),
                               GaussianNoise::generateGaussianNoise(y,VARIANCE));
            
            measurementfile << j << " " << noisy.x() << " " << noisy.y() << " " ;
        }
        
        groundtruthfile << points[i].x() << " " << points[i].y() << " " << points[i].z() << endl;
        measurementfile << endl;
        
    }
}


